/**
 * Fifo Process
 * - push/pop failure returns false
 * - it takes 51 seconds and 6GB of RAM to check that
 *   it contains no assertion violation with
 *   ~$ spin -search -bfs fifo_process_01.pml
 *
 * author: Patrick Trentin <patrick.trentin@unitn.it>
 *
 */

mtype = { PUSH, POP, IS_EMPTY, IS_FULL, RETURN };

#define PRODUCER_UID 0
#define CONSUMER_UID 1
#define FIFO_SIZE    3

proctype fifo(chan in, out)
{
    byte data[FIFO_SIZE]; // internal storage
    byte head = 0;        // points at fifo head in data
    byte count = 0;       // number of elements in fifo

    mtype command;
    byte uid, value;
    bool ret;

    do
        :: in?command(uid, value) ->
            if
                :: command == PUSH ->
                    if
                        :: count >= FIFO_SIZE ->
                            out!RETURN(uid, false, 0);
                        :: else ->
                            data[(head + count) % FIFO_SIZE] = value;
                            count = count + 1;
                            out!RETURN(uid, true, 0);
                    fi

                :: command == POP ->
                    if
                        :: count <= 0 ->
                            out!RETURN(uid, false, 0);
                        :: else ->
                            out!RETURN(uid, true, data[head]);
                            head = (head + 1) % FIFO_SIZE;
                            count = count - 1;
                    fi

                :: command == IS_EMPTY ->
                    out!RETURN(uid, count <= 0, 0);

                :: command == IS_FULL ->
                    out!RETURN(uid, FIFO_SIZE <= count, 0);
            fi;
    od;
}

proctype producer(chan fifo_in, fifo_out)
{
    mtype command;
    bool res;
    byte value;

    do
        :: atomic {
                fifo_in!IS_FULL(PRODUCER_UID, 0) ->
                fifo_out?RETURN(PRODUCER_UID, res, 0);
            }

            if
                :: res -> skip

                :: else ->
                    select(value: 0..16);
                    printf("P[%d] - produced: %d\n", _pid, value);
access_fifo:
                    atomic {
                        fifo_in!PUSH(PRODUCER_UID, value) ->
                        fifo_out?RETURN(PRODUCER_UID, res, 0);
                    }
                    assert(res);
            fi;
    od;
}

proctype consumer(chan fifo_in, fifo_out)
{
    mtype command;
	bool res;
    byte value;

    do
		:: atomic {
                fifo_in!IS_EMPTY(CONSUMER_UID, 0) ->
                fifo_out?RETURN(CONSUMER_UID, res, 0);
            }
            if
                :: res -> skip

                :: else ->
access_fifo:
                    atomic {
                        fifo_in!POP(CONSUMER_UID, 0) ->
                        fifo_out?RETURN(CONSUMER_UID, res, value);
                    }
                    assert(res);
                    printf("P[%d] - consumed: %d\n", _pid, value);
            fi;
    od;
}

init {
    chan fifo_in  = [0] of { mtype, byte, byte };
    chan fifo_out = [0] of { mtype, byte, bool, byte };

    run fifo(fifo_in, fifo_out);     // pid: 1
    run producer(fifo_in, fifo_out); // pid: 2
    run consumer(fifo_in, fifo_out); // pid: 3
}
