/**
 * Alternating Bit Protocol
 *
 * 'Fixed Code'
 *
 * Notice that writing:
 * 1A do
 * 2A     :: channel01?message(...);
 * 3A     :: channel02?message(...);
 * 4A od;
 * and writing:
 * 1B do
 * 2B     :: true -> channel01?message(...);
 * 3B     :: true -> channel02?message(...);
 * 4B od;
 * is not semantically equivalent!
 *
 * If a process executes the code in lines 1A-4A, it will block
 * at line 1A until either one of the two channels provides some message
 * to be read.
 *
 * If a process executes the code in lines 1B-4B, it will either block
 * at line 2B waiting for a message coming from channel01, or block at line
 * 3B waiting for a message coming from channel02.
 *
 *
 * NOTE(PT): the proposed solution may work in practice, but it not
 * sufficient to exclude --from a formal verification point of view--
 * a degeneration of the protocol. In fact, one can prove that
 * there exists an execution trace for which the protocol does not
 * progress and it no longer sends 1/0 messages. To do so, one
 * can verify whether there exists an infinite execution trace
 * for which at some poin the execution no longer goes through
 * the designated progress state:
 *
 * ~$ spin -a -DNP test.pml
 * ~$ gcc -DNP pan.c -o run
 * ~$ ./run -l -bfs
 * ...
 * State-vector 88 byte, depth reached 41, errors: 1
 * ...
 *
 * author: unknown
 *
 */

mtype = { MESSAGE, ACK };

chan sender2receiver = [2] of { mtype, bit, int};
chan receiver2sender = [2] of { mtype, bit, int};

inline unreliable_send(channel, type, tag, seqno) {
    bool loss = false;
    bool duplicate = true;
    if
        :: channel!type(tag, seqno);
            if
                :: channel!type(tag, seqno);
                :: duplicate = false;
            fi
        :: loss = true;
    fi
}

active proctype Sender () {
    bit in_bit, out_bit;
    int seq_no;

    // solution 1: keep sending a message till an ack is received
    do
        :: unreliable_send(sender2receiver, MESSAGE, out_bit, seq_no);
        :: receiver2sender?ACK(in_bit, 0);
            if
                :: in_bit == out_bit ->
progress:
                    out_bit = 1 - out_bit;
                    seq_no++;
                :: else ->
                    skip;
            fi;
    od;
}

active proctype Receiver () {
    bit in_bit, old_bit;
    int seq_no;

    // solution 2: keep sending acks for old messages (alternative)
    do
        :: sender2receiver?MESSAGE(in_bit, seq_no) ->
            if
                :: in_bit != old_bit ->
                    printf("received: %d\n", seq_no);
                    old_bit = in_bit;
                :: else ->
                    skip;
            fi;
         :: unreliable_send(receiver2sender, ACK, in_bit, 0);
    od;
}

/**
 * Either of the two solutions is sufficient to prevent deadlock.
 * Other solutions might exist.
 */
